/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2010 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include <net/if.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <linux/if_bridge.h>
#include "lldp.h"
#include "lldp_basman.h"
#include "messages.h"
#include "clif_msgs.h"
#include "config.h"
#include "libconfig.h"
#include "lldp_mand_clif.h"


#define SYSNAME_DEFAULT "localhost"

#define MAX_MANADDR_TLV_LENGTH	168

#define IFNUM_SUBTYPE_UNKNOWN 1
#define IFNUM_SUBTYPE_IFINDEX 2
#define IFNUM_SUBTYPE_PORTNUM 3

extern struct lldp_head lldp_head;

static const struct lldp_mod_ops basman_ops =  {
	.lldp_mod_register 	= basman_register,
	.lldp_mod_unregister 	= basman_unregister,
	.lldp_mod_gettlv	= basman_gettlv,
	.lldp_mod_rchange	= basman_rchange,
	.lldp_mod_ifup		= basman_ifup,
	.lldp_mod_ifdown	= basman_ifdown,
};

static struct basman_data *basman_data(const char *ifname)
{
	struct basman_user_data *bud;
	struct basman_data *bd = NULL;

	bud = find_module_user_data_by_if(ifname, &lldp_head, LLDP_MOD_BASIC);
	if (bud) {
		LIST_FOREACH(bd, &bud->head, entry) {
			if (!strncmp(ifname, bd->ifname, IFNAMSIZ))
				return bd;
		}
	}
	fprintf(stderr, "### %s:not found bd for %s\n", __func__, ifname);
	return NULL;
}

/*
 * basman_bld_portdesc_tlv - build port description TLV
 * @bd: the basman data struct
 * 
 * Returns 0 for success or error code for failure
 *
 * Must be called w/ init_cfg() called already
 */
static int basman_bld_portdesc_tlv(struct basman_data *bd)
{
	int length;
	int rc = 0;
	char desc[256];
	char path[256];
	char *str = NULL;
	struct unpacked_tlv *tlv = NULL;

	/* free old if it's there */
	FREE_UNPKD_TLV(bd, portdesc);

	if (!get_tlvtxenable(bd->ifname, PORT_DESCRIPTION_TLV)) {
		fprintf(stderr, "### %s:%s:Port Description disabled\n",
			__func__, bd->ifname);
		goto out_err;
	}

	/* load from config */
	memset(path, 0, sizeof(path));
	snprintf(path, sizeof(path), "tlvid%08x.info", PORT_DESCRIPTION_TLV);
	if ((get_config_setting(bd->ifname, path, (void *)&str,
	     CONFIG_TYPE_STRING)) && (str != NULL)) {
		/* use what's in the config */
		length = snprintf(desc, sizeof(desc), "%s", str); 
		fprintf(stderr, "### %s:%s:configed as %s\n",
			__func__, bd->ifname, desc);
	} else {
		length = snprintf(desc, sizeof(desc), "Interface %3d as %s",
		 	  if_nametoindex(bd->ifname), bd->ifname);
		fprintf(stderr, "### %s:%s:built as %s\n",
			__func__, bd->ifname, desc);
	}
	if (length >= sizeof(desc))
		length = sizeof(desc) - 1;

	tlv = create_tlv();
	if (!tlv)
		goto out_err;

	tlv->type = PORT_DESCRIPTION_TLV;
	tlv->length = length;
	tlv->info = (u8 *)malloc(tlv->length);
	if(!tlv->info) {
		fprintf(stderr, "### %s:%s:malloc(%d) bytes failed\n",
			__func__, bd->ifname, tlv->length);
		free(tlv);
		tlv = NULL;
		goto out_err;
	}
	memcpy(tlv->info, desc, tlv->length);
	bd->portdesc = tlv;
	rc = 0;
out_err:
	return rc;
}

/*
 * basman_bld_sysname_tlv - build port description TLV
 * @bd: the basman data struct
 * 
 * Returns 0 for success or error code for failure
 *
 * Must be called w/ init_cfg() called already
 */
static int basman_bld_sysname_tlv(struct basman_data *bd)
{
	int length;
	int rc = 0;
	char desc[256];
	char path[256];
	char *str = NULL;
	struct utsname uts;
	struct unpacked_tlv *tlv = NULL;

	/* free old if it's there */
	FREE_UNPKD_TLV(bd, sysname);

	if (!get_tlvtxenable(bd->ifname, SYSTEM_NAME_TLV)) {
		fprintf(stderr, "### %s:%s:System Name disabled\n",
			__func__, bd->ifname);
		goto out_err;
	}

	/* load from config */
	memset(path, 0, sizeof(path));
	snprintf(path, sizeof(path), "tlvid%08x.info", SYSTEM_NAME_TLV);
	if ((get_config_setting(bd->ifname, path, (void *)&str,
	     CONFIG_TYPE_STRING)) && (str != NULL)) {
		/* use what's in the config */
		length = snprintf(desc, sizeof(desc), "%s", str); 
		fprintf(stderr, "### %s:%s:configed as %s\n",
			__func__, bd->ifname, desc);
	} else {
		if (uname(&uts))
			length = snprintf(desc, sizeof(desc), SYSNAME_DEFAULT);
		else
			length = snprintf(desc, sizeof(desc), uts.nodename);
		fprintf(stderr, "### %s:%s:built as %s\n",
			__func__, bd->ifname, desc);
	}
	if (length >= sizeof(desc))
		length = sizeof(desc) - 1;

	tlv = create_tlv();
	if (!tlv)
		goto out_err;

	tlv->type = SYSTEM_NAME_TLV;
	tlv->length = length;
	tlv->info = (u8 *)malloc(tlv->length);
	if(!tlv->info){
		fprintf(stderr, "### %s:%s:malloc(%d) bytes failed\n",
			__func__, bd->ifname, tlv->length);
		free(tlv);
		tlv = NULL;
		goto out_err;
	}
	memcpy(tlv->info, desc, tlv->length);
	bd->sysname = tlv;
	rc = 0;
out_err:
	return rc;
}


/*
 * basman_bld_sysdesc_tlv - build port description TLV
 * @bd: the basman data struct
 * 
 * Returns 0 for success or error code for failure
 *
 * Must be called w/ init_cfg() called already.
 *
 * net-snmp-utils: snmptest returns the following for sysDesr: `uname-a`
 */
static int basman_bld_sysdesc_tlv(struct basman_data *bd)
{
	int length;
	int rc = 0;
	char desc[256];
	char path[256];
	char *str = NULL;
	struct utsname uts;
	struct unpacked_tlv *tlv = NULL;

	/* free old if it's there */
	FREE_UNPKD_TLV(bd, sysdesc);

	if (!get_tlvtxenable(bd->ifname, SYSTEM_DESCRIPTION_TLV)) {
		fprintf(stderr, "### %s:%s:System Description disabled\n",
			__func__, bd->ifname);
		goto out_err;
	}

	/* load from config */
	memset(path, 0, sizeof(path));
	snprintf(path, sizeof(path), "tlvid%08x.info", SYSTEM_NAME_TLV);
	if ((get_config_setting(bd->ifname, path, (void *)&str,
	     CONFIG_TYPE_STRING)) && (str != NULL)) {
		/* use what's in the config */
		length = snprintf(desc, sizeof(desc), "%s", str); 
		fprintf(stderr, "### %s:%s:configed as %s\n",
			__func__, bd->ifname, desc);
	} else {
		if (uname(&uts)) {
			length = snprintf(desc, sizeof(desc), "Unknown system");
		} else {
			length = snprintf(desc, sizeof(desc), "%s %s %s %s %s", 
					  uts.sysname, uts.nodename, uts.release,
					  uts.version, uts.machine); 
		}
		fprintf(stderr, "### %s:%s:built as %s\n",
			__func__, bd->ifname, desc);
	}
	if (length >= sizeof(desc))
		length = sizeof(desc) - 1;

	tlv = create_tlv();
	if (!tlv)
		goto out_err;

	tlv->type = SYSTEM_DESCRIPTION_TLV;
	tlv->length = length;
	tlv->info = (u8 *)malloc(tlv->length);
	if(!tlv->info){
		fprintf(stderr, "### %s:%s:malloc(%d) bytes failed\n",
			__func__, bd->ifname, tlv->length);
		free(tlv);
		tlv = NULL;
		goto out_err;
	}
	memcpy(tlv->info, desc, tlv->length);
	bd->sysdesc = tlv;
	rc = 0;
out_err:
	return rc;
}

/*
 * basman_bld_syscaps_tlv - build port description TLV
 * @bd: the basman data struct
 * 
 * Returns 0 for success or error code for failure
 *
 * Must be called w/ init_cfg() called already.
 *
 * TODO: 
 *  - This is mandatory for LLDP-MED Class III
 *  - TPID to determine C-VLAN vs. S-VLAN ?
 */
static int basman_bld_syscaps_tlv(struct basman_data *bd)
{
	int rc = 0;
	u32 syscaps;
	struct unpacked_tlv *tlv = NULL;

	/* free old if it's there */
	FREE_UNPKD_TLV(bd, syscaps);

	if (!get_tlvtxenable(bd->ifname, SYSTEM_CAPABILITIES_TLV)) {
		fprintf(stderr, "### %s:%s:System Capabilities disabled\n",
			__func__, bd->ifname);
		goto out_err;
	}

	/* load from config */
	if (get_config_tlvinfo(bd->ifname, TLVID(0, SYSTEM_CAPABILITIES_TLV),
			      (void *)&syscaps, sizeof(syscaps))) {
		u16 *caps = (u16 *)&syscaps;
		fprintf(stderr, "### %s:%s:Build System Caps from scratch\n",
			__func__, bd->ifname);
		caps[0] = htons(get_caps(bd->ifname));
		caps[1] = (is_active(bd->ifname)) ? caps[0] : 0;
	}

	tlv = create_tlv();
	if (!tlv)
		goto out_err;

	tlv->type = SYSTEM_CAPABILITIES_TLV;
	tlv->length = sizeof(syscaps);
	tlv->info = (u8 *)malloc(tlv->length);
	if(!tlv->info) {
		fprintf(stderr, "### %s:%s:malloc(%d) bytes failed\n",
			__func__, bd->ifname, tlv->length);
		free(tlv);
		tlv = NULL;
		goto out_err;
	}
	memcpy(tlv->info, &syscaps, tlv->length);
	bd->syscaps = tlv;
	rc = 0;
out_err:
	return rc;
}

static int basman_bld_maddrtlv(struct basman_data *bd, u8 masub)
{
	int rc = 0;
	int malen = 0;
	u32 ifidx;
	struct unpacked_tlv *tlv = NULL;
	union {
		struct in_addr in;
		struct in6_addr in6;
		u8 mac[6];
	} __attribute__ ((__packed__)) ma;

	if (bd->macnt == MANADDR_MAX) {
		fprintf(stderr, "### %s:%s:reached max %d Management Address\n",
			__func__, bd->ifname, bd->macnt);
		goto out_err;
	}

	switch(masub) {
	case MANADDR_IPV4:
		malen = sizeof(struct in_addr);
		if (get_ipaddr(bd->ifname, &ma.in)) {
			fprintf(stderr, "### %s:%s:IPV4 failed\n",
				__func__, bd->ifname);
			goto out_err;
		}
		break;
	case MANADDR_IPV6:
		malen = sizeof(struct in6_addr);
		if (get_ipaddr6(bd->ifname, &ma.in6)) {
			fprintf(stderr, "### %s:%s:IPV6 failed\n",
				__func__, bd->ifname);
			goto out_err;
		}
		break;
	case MANADDR_ALL802:
		masub = MANADDR_ALL802;
		malen = sizeof(ma.mac);
		if (get_mac(bd->ifname, ma.mac)) {
			fprintf(stderr, "### %s:%s:ALL802 failed\n",
				__func__, bd->ifname);
			goto out_err;
		}
		break;
	default:
		fprintf(stderr, "### %s:%s:madrr type %d not supported\n",
			__func__, bd->ifname, masub);
		goto out_err;
	}
	
	tlv = create_tlv();
	if (!tlv)
		goto out_err;

	/*
	 *  mlen + msub + maddr  + ifsub + ifidx + oidlen + oid 
	 *  1   + 1    + [1-31] + 1     + 4     + 1       + [1-128]
	 */
	tlv->length = malen + 8;
	tlv->type = MANAGEMENT_ADDRESS_TLV;
	tlv->info = (u8 *)malloc(tlv->length);
	if(!tlv->info) {
		fprintf(stderr, "### %s:%s:malloc(%d) bytes failed\n",
			__func__, bd->ifname, tlv->length);
		free(tlv);
		tlv = NULL;
		goto out_err;
	}

	/* man addr */
	tlv->info[0] = malen + 1;
	tlv->info[1] = masub;
	memcpy(&tlv->info[2], &ma, malen);
	malen += 2;

	/* interface number */
	ifidx = htonl(if_nametoindex(bd->ifname));
	tlv->info[malen] = IFNUM_SUBTYPE_IFINDEX;
	malen++;
	memcpy(&tlv->info[malen], &ifidx, sizeof(ifidx));
	malen += sizeof(ifidx);

	/* TODO: OID */
	tlv->info[malen] = 0;

	/* add to list */
	fprintf(stderr, "### %s:%s:maddr[%d]:type %d:length %d bytes \n",
		__func__, bd->ifname, bd->macnt, masub, tlv->length);
	bd->manaddr[bd->macnt] = tlv;
	bd->macnt++;
	rc = 0;
out_err:
	return rc;

}
/*
 * basman_bld_manddrr_tlv - build management address TLV
 * @bd: the basman data struct
 * 
 * Returns 0 for success or error code for failure
 *
 * Must be called w/ init_cfg() called already.
 * Use as many existing as possible
 * Preference is config > ipv6 > ipv4 > mac > default
 *
 * Max info length is = 1 + 1 + 32 + 1 + 4 + 1 + 128 = 168
 * TODO: 
 *  - No support for OID yet
 */
static int basman_bld_manaddr_tlv(struct basman_data *bd)
{
	int i;
	int rc = 0;
	struct unpacked_tlv *tlv = NULL;
	u8 ma_cfg[MAX_MANADDR_TLV_LENGTH];

	for (i = 0; i < bd->macnt; i++)
		FREE_UNPKD_TLV(bd, manaddr[i]);
	bd->macnt = 0;

	if (!get_tlvtxenable(bd->ifname, MANAGEMENT_ADDRESS_TLV)) {
		fprintf(stderr, "### %s:%s:Management Address disabled\n",
			__func__, bd->ifname);
		goto out_err;
	}

	/* load from config */
	if (!get_config_tlvinfo(bd->ifname, TLVID(0, SYSTEM_CAPABILITIES_TLV),
			      (void *)ma_cfg, sizeof(ma_cfg))) {
		fprintf(stderr, "### %s:%s:Build Management Addr from config\n",
			__func__, bd->ifname);
		goto out_err;
	}
	/* build from scratch */
	basman_bld_maddrtlv(bd, MANADDR_IPV4);
	basman_bld_maddrtlv(bd, MANADDR_IPV6);
	basman_bld_maddrtlv(bd, MANADDR_ALL802);
	rc = 0;
out_err:
	return rc;
}

static void basman_free_tlv(struct basman_data *bd)
{
	int i = 0;

	if (bd) {
		FREE_UNPKD_TLV(bd, portdesc);
		FREE_UNPKD_TLV(bd, sysname);
		FREE_UNPKD_TLV(bd, sysdesc);
		FREE_UNPKD_TLV(bd, syscaps);
		for (i = 0; i < bd->macnt; i++)
			FREE_UNPKD_TLV(bd, manaddr[i]);
		bd->macnt = 0;
	}
}

/* build unpacked tlvs */
static int basman_bld_tlv(struct basman_data *bd)
{
	int rc = EPERM;

	if (!port_find_by_name(bd->ifname)) {
		rc = EEXIST;
		goto out_err;
	}

	if (!init_cfg())
		goto out_err;

	if (basman_bld_portdesc_tlv(bd)) {
		fprintf(stderr, "### %s:%s:basman_bld_portdesc_tlv() failed\n",
				__func__, bd->ifname);
		goto out_err_destroy;
	}
	if (basman_bld_sysname_tlv(bd)) {
		fprintf(stderr, "### %s:%s:basman_bld_sysname_tlv() failed\n",
				__func__, bd->ifname);
		goto out_err_destroy;
	}
	if (basman_bld_sysdesc_tlv(bd)) {
		fprintf(stderr, "### %s:%s:basman_bld_sysdesc_tlv() failed\n",
				__func__, bd->ifname);
		goto out_err_destroy;
	}
	if (basman_bld_syscaps_tlv(bd)) {
		fprintf(stderr, "### %s:%s:basman_bld_syscaps_tlv() failed\n",
				__func__, bd->ifname);
		goto out_err_destroy;
	}
	if (basman_bld_manaddr_tlv(bd)) {
		fprintf(stderr, "### %s:%s:basman_bld_manaddr_tlv() failed\n",
				__func__, bd->ifname);
		goto out_err_destroy;
	}
	rc = 0;

out_err_destroy:
	destroy_cfg();
out_err:
	return rc;
}

static void basman_free_data(struct basman_user_data *bud)
{
	struct basman_data *bd;
	if (bud) {
		while (!LIST_EMPTY(&bud->head)) {
			bd = LIST_FIRST(&bud->head);
			LIST_REMOVE(bd, entry);
			basman_free_tlv(bd);
			free(bd);
 		}
	}
}

struct packed_tlv *basman_gettlv(struct port *port)
{
	int i;
	int size;
	struct basman_data *bd;
	struct packed_tlv *ptlv = NULL;

	bd = basman_data(port->ifname);
	if (!bd)
		goto out_err;

	/* free and rebuild the TLVs */
	basman_free_tlv(bd);
	if (basman_bld_tlv(bd)) {
		fprintf(stderr, "### %s:%s basman_bld_tlv failed\n",
			__func__, port->ifname); 
		goto out_err;
	}

	ptlv = (struct packed_tlv *)malloc(sizeof(*ptlv));
	if (!ptlv) {
		fprintf(stderr, "### %s:%s malloc(ptlv) failed\n",
			__func__, port->ifname); 
		goto out_err;
	}

	size = ((bd->portdesc) ? (2 + bd->portdesc->length) : 0)
		+ ((bd->sysname) ? (2 + bd->sysname->length) : 0)
		+ ((bd->sysdesc) ? (2 + bd->sysdesc->length) : 0)
		+ ((bd->syscaps) ? (2 + bd->syscaps->length) : 0);

	for (i = 0; i < bd->macnt; i++)
		size += ((bd->manaddr[i]) ? (2 + bd->manaddr[i]->length) : 0);

	if (!size)
		goto out_err;

	ptlv->tlv = malloc(size);
	if (!ptlv->tlv)
		goto out_free;
	
	/* pack all pre-built tlvs */
	ptlv->size = 0;
	PACK_TLV_AFTER(bd->portdesc, ptlv, size, out_free);
	PACK_TLV_AFTER(bd->sysname, ptlv, size, out_free);
	PACK_TLV_AFTER(bd->sysdesc, ptlv, size, out_free);
	PACK_TLV_AFTER(bd->syscaps, ptlv, size, out_free);
	for (i = 0; i < bd->macnt; i++)
		PACK_TLV_AFTER(bd->manaddr[i], ptlv, size, out_free);

	fprintf(stderr, "### %s:%s: size=%d\n", __func__, port->ifname, ptlv->size);
	return ptlv;
out_free:
	ptlv = free_pkd_tlv(ptlv);
out_err:
	fprintf(stderr,"### %s:%s: failed\n", __func__, port->ifname);
	return NULL;

}


void basman_ifdown(char *ifname)
{
	struct basman_data *bd;

	bd = basman_data(ifname);
	if (!bd)
		goto out_err;

	LIST_REMOVE(bd, entry);
	basman_free_tlv(bd);
	free(bd);
	fprintf(stderr, "### %s:port %s removed\n", __func__, ifname); 
	return;
out_err:
	fprintf(stderr, "### %s:port %s adding failed\n", __func__, ifname); 
	return;
}

void basman_ifup(char *ifname)
{
	struct basman_data *bd;
	struct basman_user_data *bud;

	bd = basman_data(ifname);
	if (bd) {
		fprintf(stderr, "### %s:%s exists\n", __func__, ifname); 
		goto out_err;
	}

	/* not found, alloc/init per-port tlv data */
	bd = (struct basman_data *) malloc(sizeof(*bd));
	if (!bd) {
		fprintf(stderr, "### %s:%s malloc %d failed\n", __func__, ifname, sizeof(*bd)); 
		goto out_err;
	}
	memset(bd, 0, sizeof(struct basman_data));
	strncpy(bd->ifname, ifname, IFNAMSIZ);
	if (basman_bld_tlv(bd)) {
		fprintf(stderr, "### %s:%s mand_bld_tlv failed\n", __func__, ifname); 
		free(bd);
		goto out_err;
	}

	bud = find_module_user_data_by_if(ifname, &lldp_head, LLDP_MOD_BASIC); 
	LIST_INSERT_HEAD(&bud->head, bd, entry);
	fprintf(stderr, "### %s:port %s added\n", __func__, ifname); 
	return;
out_err:
	fprintf(stderr, "### %s:port %s adding failed\n", __func__, ifname); 
	return;
}

struct lldp_module *basman_register(void)
{
	struct lldp_module *mod;
	struct basman_user_data *bud;

	mod = malloc(sizeof(*mod));
	if (!mod) {
		fprintf(stderr, "failed to malloc module data\n");
		log_message(MSG_ERR_SERVICE_START_FAILURE,
			"%s", "failed to malloc module data");
		goto out_err;
	}
	bud = malloc(sizeof(struct basman_user_data));
	if (!bud) {
		free(mod);
		fprintf(stderr, "failed to malloc module user data\n");
		log_message(MSG_ERR_SERVICE_START_FAILURE,
			"%s", "failed to malloc module user data");
		goto out_err;
	}
	LIST_INIT(&bud->head);
 	mod->id = LLDP_MOD_BASIC;
	mod->ops = &basman_ops;
	mod->data = bud;

	fprintf(stderr, "### %s:done\n", __func__);
	return mod;
out_err:
	fprintf(stderr, "### %s:failed\n", __func__);
	return;

}

void basman_unregister(struct lldp_module *mod)
{
	struct basman_user_data *bud;

	if (mod->data) {
		basman_free_data((struct basman_user_data *) mod->data);
		free(mod->data);
	}
	free(mod);
	fprintf(stderr, "### %s:done\n", __func__);
}

int basman_rchange(struct port *port,  struct unpacked_tlv *tlv)
{
	return 0;
}
