/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2010 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "includes.h"
#include "common.h"
#include <stdio.h>
#include <syslog.h>
#include <sys/un.h>
#include <sys/stat.h>
#include "lldp_mod.h"
#include "lldptool.h"
#include "lldp_basman.h"
#include "lldp_basman_clif.h"

void print_string_tlv(u16, char *info);
void print_capabilities(u16, char *info);
void print_mng_addr(u16, char *info);
u32 basman_lookup_tlv_name(char *tlvid_str);
int basman_print_help();

static const struct lldp_mod_ops basman_ops_clif = {
	.lldp_mod_register 	= basman_cli_register,
	.lldp_mod_unregister 	= basman_cli_unregister,
	.print_tlv		= basman_print_tlv,
	.lookup_tlv_name	= basman_lookup_tlv_name,
	.print_help		= basman_print_help,
};

struct type_name_info basman_tlv_names[] = {
	{ 4, "Port Description TLV",    "portDesc", print_string_tlv },
	{ 5, "System Name TLV",         "sysName",  print_string_tlv },
	{ 6, "System Description TLV",  "sysDesc",  print_string_tlv },
	{ 7, "System Capabilities TLV", "sysCap",   print_capabilities },
	{ 8, "Management Address TLV",  "mngAddr",  print_mng_addr },
	{ INVALID_TLVID,	NULL, NULL }
};

int basman_print_help()
{
	int i;
	struct type_name_info *tn = &basman_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (tn->key && strlen(tn->key) && tn->name) {
			printf("   %s", tn->key);
			if (strlen(tn->key)+3 <= 8)
				printf("\t");
			printf("\t: %s\n", tn->name);
		}
		tn++;
	}
	
	return 0;
}

struct lldp_module *basman_cli_register(void)
{
	struct lldp_module *mod;

	mod = malloc(sizeof(*mod));
	if (!mod) {
		fprintf(stderr, "failed to malloc module data\n");
		goto out_err;
	}
 	mod->id = LLDP_MOD_BASIC;
	mod->ops = &basman_ops_clif;

	return mod;
out_err:
	return NULL;
}

void basman_cli_unregister(struct lldp_module *mod)
{
	free(mod);
}

void print_string_tlv(u16 len, char *info)
{
	int i;
	for (i = 0; i < 2*len; i+=2)
		printf("%c", hex2int(info+i));
	printf("\n");
}

void print_capabilities(u16 len, char *info)
{
	printf("system capabilities: %*.*s", len, len, info);
	printf("  enabled capabilities: %*.*s\n", len, len, info+len);
}

void print_mng_addr(u16 len, char *info)
{
	printf("%*.*s\n", 2*len, 2*len, info);
}

/* return 1: if it printed the TLV
 *        0: if it did not
 */
int basman_print_tlv(u32 tlvid, u16 len, u8 *info)
{
	int i;
	struct type_name_info *tn = &basman_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (tlvid == tn->type) {
			printf("%s\n\t", tn->name);
			if (tn->print_info)
				tn->print_info(len, info);
			return 1;
		}
		tn++;
	}
	
	return 0;
}

u32 basman_lookup_tlv_name(char *tlvid_str)
{
	struct type_name_info *tn = &basman_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (!strcasecmp(tn->key, tlvid_str))
			return tn->type;
		tn++;
	}
	return INVALID_TLVID;
}
