/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2010 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include <net/if.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include "includes.h"
#include "common.h"
#include "ctrl_iface.h"
#include "lldp.h"
#include "lldp_mand.h"
#include "lldp_mand_cmds.h"
#include "messages.h"

extern struct lldp_head lldp_head;

static const struct lldp_mod_ops mand_ops = {
	.lldp_mod_register 	= mand_register,
	.lldp_mod_unregister 	= mand_unregister,
	.lldp_mod_gettlv	= mand_gettlv,
	.lldp_mod_rchange	= mand_rchange,
	.lldp_mod_ifup		= mand_ifup,
	.lldp_mod_ifdown	= mand_ifdown,
	.client_cmd		= mand_clif_cmd,
};

static struct mand_data *mand_data(const char *ifname)
{
	struct mand_user_data *mud;
	struct mand_data *md = NULL;

	mud = find_module_user_data_by_if(ifname, &lldp_head, LLDP_MOD_MAND);
	if (mud) {
		LIST_FOREACH(md, &mud->head, entry) {
			if (!strncmp(ifname, md->ifname, IFNAMSIZ))
				return md;
		}
	}
	fprintf(stderr, "### %s:not found md for %s\n", __func__, ifname);
	return NULL;
}

static struct unpacked_tlv *mand_bld_end_tlv()
{
	struct unpacked_tlv *tlv = create_tlv();

	if(tlv) {
		tlv->type = END_OF_LLDPDU_TLV;
		tlv->length = 0;
		tlv->info = NULL;
	}
	return tlv;
}

/* TODO:
 *  - Mandatory for LLDP-MED Network Connectivity w/ default to MAC
 *  - Mandatory for LLDP-MED Endpoint w/ default to network addr
 */
static struct unpacked_tlv *mand_bld_chassis_tlv(struct port *port)
{
	u8 mac[6];
	u8 subtype;
	int length;
	struct unpacked_tlv *tlv = NULL;

	tlv = create_tlv();
	if (!tlv) {
		fprintf(stderr, "### %s:%s:create_tvl() failed\n", __func__, port->ifname);
		return NULL;
	}
	/* MAC is preferred */
	get_mac(port->ifname, mac);
	if (is_valid_mac(mac)) {
		subtype = CHASSIS_ID_MAC_ADDRESS;
		length = sizeof(mac) + sizeof(subtype);
	} else {
		subtype = CHASSIS_ID_INTERFACE_NAME;
		length = (u16)(strlen(port->ifname)) + sizeof(subtype);
	}
	tlv->type = CHASSIS_ID_TLV;
	tlv->length = length;
	tlv->info = (u8 *)malloc(length);
	if(!tlv->info) {
		fprintf(stderr, "### %s:%s:malloc() failed\n", __func__, port->ifname);
		free(tlv);
		return NULL;
	}

	memset(tlv->info, 0, tlv->length);
	tlv->info[0] = subtype;
	if (subtype == CHASSIS_ID_MAC_ADDRESS)
		memcpy(&tlv->info[1], mac, tlv->length - 1);
	else
		memcpy(&tlv->info[1], port->ifname, tlv->length - 1);

	fprintf(stderr, "### %s:%s:done:type=%d length=%d\n", __func__,
		port->ifname, tlv->type, tlv->length);

	return tlv;
}

static struct unpacked_tlv *mand_bld_portid_tlv(struct port *port)
{
	struct unpacked_tlv *tlv = create_tlv();
	u8 own_addr[ETH_ALEN], *hwaddr;

	if (!tlv)
		return NULL;

	hwaddr = &own_addr[0];
	tlv->type = PORT_ID_TLV;
	tlv->length = 7;
	tlv->info = (u8 *)malloc(tlv->length);
	if(tlv->info){
		memset(tlv->info,0, tlv->length);
		tlv->info[0] = PORT_ID_MAC_ADDRESS;
		if (l2_packet_get_own_addr(port->l2, hwaddr) >= 0)
			memcpy(&tlv->info[1], &own_addr, ETH_ALEN);
		fprintf(stderr, "### %s:%s:done:type=%d length=%d\n", __func__,
			port->ifname, tlv->type, tlv->length);
	} else {
		fprintf(stderr, "### %s:%s:failed\n", __func__, port->ifname);
		free(tlv);
		return NULL;
	}
	return tlv;
}

static struct unpacked_tlv *mand_bld_ttl_tlv(struct port *port)
{
	u16 ttl;
	
	struct unpacked_tlv *tlv = create_tlv();

	if (!tlv)
		return NULL;

	tlv->type = TIME_TO_LIVE_TLV;
	tlv->length = 2;
	tlv->info = (u8 *)malloc(tlv->length);
	if(tlv->info) {
		memset(tlv->info, 0, tlv->length);

		if (port->tx.txTTL)
			ttl = htons(port->tx.txTTL);
		else
			ttl = htons(DEFAULT_TX_HOLD * DEFAULT_TX_INTERVAL);

		memcpy(tlv->info, &ttl, tlv->length);
		fprintf(stderr, "### %s:%s:done:type=%d length=%d ttl=%d\n", __func__,
			port->ifname, tlv->type, tlv->length, ntohs(ttl));
	} else {
		fprintf(stderr, "### %s:%s:failed\n", __func__, port->ifname);
		free(tlv);
		return NULL;
	}
	return tlv;
}

struct packed_tlv *mand_gettlv(struct port *port)
{
	struct mand_data *md;
	struct packed_tlv *ptlv = NULL;
	int length;

	md = mand_data(port->ifname);
	if (!md) {
		fprintf(stderr, "### %s:%s: not found port\n", __func__, port->ifname);
		goto out_err;
	}

	ptlv = (struct packed_tlv *)malloc(sizeof(*ptlv));
	if (!ptlv)
		goto out_err;

	length = 255 + 255 + 2;
	ptlv->tlv = malloc(length);
	if (!ptlv->tlv)
		goto out_free;

	ptlv->size = 0;
	if (pack_tlv_after(md->chassis, ptlv, length))
		goto out_free;

	if (pack_tlv_after(md->portid, ptlv, length))
		goto out_free;

	if (pack_tlv_after(md->ttl, ptlv, length))
		goto out_free;

	fprintf(stderr, "### %s:%s: size=%d\n", __func__, port->ifname, ptlv->size);
	return ptlv;
out_free:
	ptlv = free_pkd_tlv(ptlv);
out_err:

	fprintf(stderr,"### %s:%s: failed\n", __func__, port->ifname);
	return NULL;

}

static void mand_free_tlv(struct mand_data *md)
{
	if (md) {
		FREE_UNPKD_TLV(md, chassis);
		FREE_UNPKD_TLV(md, portid);
		FREE_UNPKD_TLV(md, ttl);
		FREE_UNPKD_TLV(md, end);
	}
}

/* build unpacked tlvs */
static int mand_bld_tlv(struct mand_data *md)
{
	struct port *port;

	port = port_find_by_name(md->ifname);
	if (!port) {
		fprintf(stderr, "### %s: failed for %s\n", __func__, md->ifname);
		return EEXIST;
	}
	if (!md->chassis)
		md->chassis = mand_bld_chassis_tlv(port);
	if (!md->portid)
		md->portid = mand_bld_portid_tlv(port);
	if (!md->ttl)
		md->ttl = mand_bld_ttl_tlv(port);
	if (!md->end)
		md->end = mand_bld_end_tlv(port);

	return 0;

}

static void mand_free_data(struct mand_user_data *mud)
{
	struct mand_data *md;
	if (mud) {
		while (!LIST_EMPTY(&mud->head)) {
			md = LIST_FIRST(&mud->head);
			LIST_REMOVE(md, entry);
			mand_free_tlv(md);
			free(md);
 		}
	}
}

void mand_ifdown(char *ifname)
{
	struct mand_data *md;

	md = mand_data(ifname);
	if (!md)
		goto out_err;

	LIST_REMOVE(md, entry);
	mand_free_tlv(md);
	free(md);
	fprintf(stderr, "### %s:port %s removed\n", __func__, ifname); 
	return;
out_err:
	fprintf(stderr, "### %s:port %s adding failed\n", __func__, ifname); 
	return;
}

void mand_ifup(char *ifname)
{
	struct mand_data *md;
	struct mand_user_data *mud;

	md = mand_data(ifname);
	if (md) {
		fprintf(stderr, "### %s:%s exists\n", __func__, ifname); 
		goto out_err;
	}
	/* not found, alloc/init per-port tlv data */
	md = (struct mand_data *) malloc(sizeof(*md));
	if (!md) {
		fprintf(stderr, "### %s:%s malloc %d failed\n", __func__, ifname, sizeof(*md)); 
		goto out_err;
	}
	memset(md, 0, sizeof(struct mand_data));
	strncpy(md->ifname, ifname, IFNAMSIZ);
	if (mand_bld_tlv(md)) {
		fprintf(stderr, "### %s:%s mand_bld_tlv failed\n", __func__, ifname); 
		free(md);
		goto out_err;
	}

	mud = find_module_user_data_by_if(ifname, &lldp_head, LLDP_MOD_MAND);
	LIST_INSERT_HEAD(&mud->head, md, entry);
	fprintf(stderr, "### %s:port %s added\n", __func__, ifname); 
	return;
out_err:
	fprintf(stderr, "### %s:port %s adding failed\n", __func__, ifname); 
	return;
}

struct lldp_module *mand_register(void)
{
	struct lldp_module *mod;
	struct mand_user_data *mud;

	mod = malloc(sizeof(*mod));
	if (!mod) {
		fprintf(stderr, "failed to malloc module data\n");
		log_message(MSG_ERR_SERVICE_START_FAILURE,
			"%s", "failed to malloc module data");
		goto out_err;
	}
	mud = malloc(sizeof(struct mand_user_data));
	if (!mud) {
		free(mod);
		fprintf(stderr, "failed to malloc module user data\n");
		log_message(MSG_ERR_SERVICE_START_FAILURE,
			"%s", "failed to malloc module user data");
		goto out_err;
	}
	LIST_INIT(&mud->head);
 	mod->id = LLDP_MOD_MAND;
	mod->ops = &mand_ops;
	mod->data = mud;

	fprintf(stderr, "### %s:done\n", __func__);
	return mod;
out_err:
	fprintf(stderr, "### %s:failed\n", __func__);
	return;
}

void mand_unregister(struct lldp_module *mod)
{
	struct mand_user_data *mud;

	if (mod->data) {
		mand_free_data((struct mand_user_data *) mod->data);
		free(mod->data);
	}

	free(mod);
	fprintf(stderr, "### %s:done\n", __func__); 
}

int mand_rchange(struct port *port,  struct unpacked_tlv *tlv)
{
	return 0;
}

