/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2010 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "includes.h"
#include "common.h"
#include <stdio.h>
#include <syslog.h>
#include <sys/un.h>
#include <sys/stat.h>
#include "lldpad.h"
#include "ctrl_iface.h"
#include "lldp_mand.h"
#include "lldp_mand_clif.h"
#include "lldp/ports.h"
#include "libconfig.h"
#include "config.h"
#include "clif_msgs.h"
#include "lldp/states.h"

int get_arg_adminstatus(struct cmd *cmd, char *arg, char *argvalue, char *obuf);
int set_arg_adminstatus(struct cmd *cmd, char *arg, char *argvalue, char *obuf);
int get_arg_tlvtxenable(struct cmd *cmd, char *arg, char *argvalue, char *obuf);
int set_arg_tlvtxenable(struct cmd *cmd, char *arg, char *argvalue, char *obuf);
int handle_get_arg(struct cmd *cmd, char *arg, char *argvalue, char *obuf);
int handle_set_arg(struct cmd *cmd, char *arg, char *argvalue, char *obuf);

static char *hexlist = "0123456789abcdef";
static char *hexlistcaps = "0123456789ABCDEF";

struct arg_handlers get_arg_handlers[] = {
	{ ARG_ADMINSTATUS,    get_arg_adminstatus },
	{ ARG_TLVTXENABLE,    get_arg_tlvtxenable },
	{ NULL }
};

struct arg_handlers set_arg_handlers[] = {
	{ ARG_ADMINSTATUS,    set_arg_adminstatus },
	{ ARG_TLVTXENABLE,    set_arg_tlvtxenable },
	{ NULL }
};

int get_arg_adminstatus(struct cmd *cmd, char *arg, char *argvalue, char *obuf)
{
	int value;
	char *s;

	if (!init_cfg())
		return cmd_failed;

	if (!get_config_setting(cmd->ifname, arg, (void *)&value,
				CONFIG_TYPE_INT))
		value = disabled;

	switch (value) {
	case disabled:
		s = VAL_DISABLED;
		break;
	case enabledTxOnly:
		s = VAL_TX;
		break;
	case enabledRxOnly:
		s = VAL_RX;
		break;
	case enabledRxTx:
		s = VAL_RXTX;
		break;
	default:
		s = VAL_INVALID;
		return cmd_invalid;
	}
	
	sprintf(obuf, "%02x%s%04x%s", strlen(arg), arg, strlen(s), s);
	destroy_cfg();
	return cmd_success;
}

int get_arg_tlvtxenable(struct cmd *cmd, char *arg, char *argvalue, char *obuf)
{
	int value;
	char *s;
	char arg_path[256];

	if (!init_cfg())
		return cmd_failed;

	snprintf(arg_path, sizeof(arg_path), "tlvid%08x.%s", cmd->tlvid, arg);
	
	if (!get_config_setting(cmd->ifname, arg_path, (void *)&value,
	 			CONFIG_TYPE_BOOL))
		value = FALSE;

	if (value)
		s = VAL_YES;
	else
		s = VAL_NO;
	
	sprintf(obuf, "%02x%s%04x%s", strlen(arg), arg, strlen(s), s);
	destroy_cfg();
	return cmd_success;
}

int handle_get_arg(struct cmd *cmd, char *arg, char *argvalue, char *obuf)
{
	struct arg_handlers *ah = &get_arg_handlers[0];

	while (ah->arg) {
		if (!strcasecmp(ah->arg, arg))
			return ah->handle_arg(cmd, ah->arg, argvalue, obuf);
		ah++;
	}
	return cmd_invalid;
}

int set_arg_adminstatus(struct cmd *cmd, char *arg, char *argvalue, char *obuf)
{
	long value;

	if (!strcasecmp(argvalue, VAL_RXTX))
		value =  enabledRxTx;
	else if (!strcasecmp(argvalue, VAL_RX))
		value = enabledRxOnly;
	else if (!strcasecmp(argvalue, VAL_TX))
		value = enabledTxOnly;
	else if (!strcasecmp(argvalue, VAL_DISABLED))
		value = disabled;
	else
		return cmd_invalid;  /* ignore invalid value */

	if (!init_cfg())
		return cmd_failed;

	if (set_config_setting(cmd->ifname, arg, (void *)&value,
			       CONFIG_TYPE_INT)) {
		return cmd_failed;
	}

	set_lldp_port_admin(cmd->ifname, value);

	destroy_cfg();

	return cmd_success;
}

int set_arg_tlvtxenable(struct cmd *cmd, char *arg, char *argvalue, char *obuf)
{
	int value;
	char arg_path[256];


	if (!strcasecmp(argvalue, VAL_YES))
		value = 1;
	else if (!strcasecmp(argvalue, VAL_NO))
		value = 0;
	else
		return cmd_invalid;  /* ignore invalid value */

	snprintf(arg_path, sizeof(arg_path), "tlvid%08x.%s", cmd->tlvid, arg);

	if (set_cfg(cmd->ifname, arg_path, (void *)&value, CONFIG_TYPE_BOOL))
		return cmd_failed;

	process_somethingChangedLocal(cmd->ifname, 0);

	return cmd_success;
}

int handle_set_arg(struct cmd *cmd, char *arg, char *argvalue, char *obuf)
{
	struct arg_handlers *ah = &set_arg_handlers[0];

	while (ah->arg) {
		if (!strcasecmp(ah->arg, arg))
			return ah->handle_arg(cmd, ah->arg, argvalue, obuf);
		ah++;
	}
	return cmd_invalid;
}


int get_tlvs(struct cmd *cmd, char *rbuf)
{
	u8 tlvs[2048];
	int size = 0;
	int i;
	u32 tlvid;
	int off = 0;
	int moff = 0;
	u16 type, len;

	if (cmd->ops & op_local) {
		if (get_local_tlvs(cmd->ifname, &tlvs[0], &size))
			return cmd_failed;
	} else if (cmd->ops & op_neighbor) {
		if (get_neighbor_tlvs(cmd->ifname, &tlvs[0], &size))
			return cmd_failed;
	} else
		return cmd_failed;

	/* filter down response if a specific TLVID was requested */
	if (cmd->tlvid != INVALID_TLVID) {
		/* step through PDU buffer and move matching TLVs to front */
		while (off < size) {
			type = *((u16 *) (tlvs+off));	
			type = htons(type);
			len = type & 0x01ff;
			type >>= 9;
			if (type < INVALID_TLVID) {
				tlvid = type;
			} else {
				tlvid = *((u32 *)(tlvs+off+2));	
				tlvid = ntohl(tlvid);
			}

			if (tlvid == cmd->tlvid) {
				memcpy(tlvs+moff, tlvs+off, sizeof(u16)+len);
				moff += sizeof(u16)+len;
			}

			off += (sizeof(u16)+len);
		}
		size = moff;
	}

	for (i = 0; i < size; i++) {
		sprintf(rbuf + 2*i, "%02x", tlvs[i]);
	}
	return cmd_success;
}

int get_port_stats(struct cmd *cmd, char *rbuf)
{
	int offset=0;
	struct portstats stats;

	if (get_lldp_port_statistics(cmd->ifname, &stats))
		return cmd_device_not_found;

	sprintf(rbuf+offset, "%08x", stats.statsFramesOutTotal);
	offset+=8;
	sprintf(rbuf+offset, "%08x", stats.statsFramesDiscardedTotal);
	offset+=8;
	sprintf(rbuf+offset, "%08x", stats.statsFramesInErrorsTotal);
	offset+=8;
	sprintf(rbuf+offset, "%08x", stats.statsFramesInTotal);
	offset+=8;
	sprintf(rbuf+offset, "%08x", stats.statsTLVsDiscardedTotal);
	offset+=8;
	sprintf(rbuf+offset, "%08x", stats.statsTLVsUnrecognizedTotal);
	offset+=8;
	sprintf(rbuf+offset, "%08x", stats.statsAgeoutsTotal);

	return cmd_success;
}

int mand_clif_cmd(void  *data,
		  struct sockaddr_un *from,
		  socklen_t fromlen,
		  char *ibuf, int ilen,
		  char *rbuf)
{
	struct cmd cmd;
	u8 len;
	u8 arglen;
	u16 argvalue_len;
	int ioff, roff;
	int rstatus = cmd_success;
	char *arg = NULL;
	char *argvalue = NULL;

	/* pull out the command elements of the command message */
	hexstr2bin(ibuf+CMD_CODE, (u8 *)&cmd.cmd, sizeof(cmd.cmd));
	hexstr2bin(ibuf+CMD_OPS, (u8 *)&cmd.ops, sizeof(cmd.ops));
	cmd.ops = ntohl(cmd.ops);
	hexstr2bin(ibuf+CMD_IF_LEN, &len, sizeof(len));
	ioff = CMD_IF;
	if (len < sizeof(cmd.ifname))
		memcpy(cmd.ifname, ibuf+CMD_IF, len);
	else
		return 1;

	cmd.ifname[len] = '\0';
	ioff += len;

	if (cmd.cmd == cmd_gettlv || cmd.cmd == cmd_settlv) {
		hexstr2bin(ibuf+ioff, (u8 *)&cmd.tlvid, sizeof(cmd.tlvid));
		cmd.tlvid = ntohl(cmd.tlvid);
		ioff += 2*sizeof(cmd.tlvid);
	}

	/* check for an arg and argvalue */
	if (ilen - ioff > 2*sizeof(arglen)) {
		hexstr2bin(ibuf+ioff, &arglen, sizeof(arglen));
		ioff += 2*sizeof(arglen);
		if (ilen - ioff >= arglen) {
			arg = ibuf+ioff;
			ioff += arglen;

			if (ilen - ioff > 2*sizeof(argvalue_len)) {
				hexstr2bin(ibuf+ioff, (u8 *)&argvalue_len,
					   sizeof(argvalue_len));
				argvalue_len = ntohs(argvalue_len);
				ioff += 2*sizeof(argvalue_len);
				if (ilen - ioff >= argvalue_len) {
					argvalue = ibuf+ioff;
					ioff += argvalue_len;
				}
			}
		}
	}

	if (arg)
		arg[arglen] = '\0';
	if (argvalue)
		argvalue[argvalue_len] = '\0';
	
	sprintf(rbuf, "%c%1x%02x%08x%02x%s", CMD_REQUEST, CLIF_MSG_VERSION,
		cmd.cmd, cmd.ops, strlen(cmd.ifname), cmd.ifname);
	roff = strlen(rbuf);

	switch (cmd.cmd) {
	case cmd_getstats:
		rstatus = get_port_stats(&cmd, rbuf+roff);
		break;
	case cmd_gettlv:
		sprintf(rbuf+roff, "%08x", cmd.tlvid);
		roff+=8;
		if (arg)
			rstatus = handle_get_arg(&cmd, arg, NULL,
						 rbuf+strlen(rbuf));
		else
			rstatus = get_tlvs(&cmd, rbuf+roff);
		break;
	case cmd_settlv:
		sprintf(rbuf+roff, "%08x", cmd.tlvid);
		roff+=8;
		rstatus = handle_set_arg(&cmd, arg, argvalue,
					 rbuf+strlen(rbuf));
		break;
	case cmd_get_lldp:
		rstatus = handle_get_arg(&cmd, arg, NULL,
					 rbuf+strlen(rbuf));
		break;
	case cmd_set_lldp:
		rstatus = handle_set_arg(&cmd, arg, argvalue,
			rbuf+strlen(rbuf));
		break;
	default:
		return cmd_invalid;
	}

	return rstatus;
}
